// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TrueTypeAsset.hpp"
#include "asset/AssetManager.hpp"

using namespace Dawn;

TrueTypeAsset::TrueTypeAsset(AssetManager *assMan, std::string name) :
  Asset(assMan, name),
  loader(name + ".truetype")
{

}

void TrueTypeAsset::updateSync() {

}

void TrueTypeAsset::updateAsync() {
  if(this->state != TRUE_TYPE_ASSET_STATE_INITIAL) return;

  this->state = TRUE_TYPE_ASSET_STATE_OPEN;
  this->loader.open();

  this->state = TRUE_TYPE_ASSET_STATE_VALIDATE_HEADER;

  uint8_t buffer[64];
  this->loader.rewind();
  size_t read = this->loader.read(buffer, sizeof(char) * 6);
  assertTrue(read == (6 * sizeof(char)));
  buffer[6] = '\0';

  // Confirm "DE_TTF"
  assertTrue(std::string((char *)buffer) == "DE_TTF");

  // Vertical bar
  this->loader.read(buffer, 1);
  assertTrue(buffer[0] == '|');

  // Read version
  this->state = TRUE_TYPE_ASSET_STATE_VALIDATE_VERSION;
  read = this->loader.read(buffer, sizeof(char) * 5);
  assertTrue(buffer[0] == '3');
  assertTrue(buffer[1] == '.');
  assertTrue(buffer[2] == '0');
  assertTrue(buffer[3] == '0');
  assertTrue(buffer[4] == '|');

  // Read the count of font styles / variants.
  size_t styleListBegin = this->loader.getPosition();
  this->state = TRUE_TYPE_ASSET_STATE_READ_VARIANT_COUNT;
  read = this->loader.read(buffer, 64);
  assertTrue(read > 0);

  // Get position of vertical bar.
  size_t i = 0;
  while(buffer[i] != '|' && i < 64) i++;
  assertTrue(buffer[i] == '|');
  styleListBegin += i + 1;
  buffer[i] = '\0';

  int32_t count = atoi((char *)buffer);
  assertTrue(count > 0);

  // Now begin parsing each font style.
  this->state = TRUE_TYPE_ASSET_STATE_READ_VARIANT;
  assetStyles.clear();
  while(assetStyles.size() != count) {
    struct TrueTypeAssetStyle style;

    // Buffer
    this->loader.rewind();
    this->loader.setPosition(styleListBegin);
    read = this->loader.read(buffer, 32);
    assertTrue(read == 32);

    // Read style
    i = 0;
    while(buffer[i] != ':' && i < 64) i++;
    buffer[i] = '\0';
    style.style = atoi((char *)buffer);
    styleListBegin += i + 1;
    
    // Buffer
    this->loader.rewind();
    this->loader.setPosition(styleListBegin);
    read = this->loader.read(buffer, 32);
    assertTrue(read == 32);

    // Read length
    i = 0;
    while(buffer[i] != '|' && i < 64) i++;
    buffer[i] = '\0';
    styleListBegin += i + 1;
    style.dataSize = atol((char *)buffer);

    // Push
    assetStyles.push_back(style);
  }

  // Now we are at the first byte of the first style.
  this->state = TRUE_TYPE_ASSET_STATE_ADJUSTING_OFFSETS;
  auto itStyle = assetStyles.begin();
  while(itStyle != assetStyles.end()) {
    (*itStyle).dataOffset = styleListBegin;
    styleListBegin += (*itStyle).dataSize;
    styleListBegin += 1; // Vertical bar
    ++itStyle;
  }

  // Init FreeType
  this->state = TRUE_TYPE_ASSET_STATE_INIT_FREETYPE;
  int32_t ret = FT_Init_FreeType(&fontLibrary);
  assertTrue(ret == 0);

  // Done parsing!
  this->state = TRUE_TYPE_ASSET_STATE_READY;
  this->loaded = true;
}

usagelockid_t TrueTypeAsset::lock(struct TrueTypeFaceTextureStyle style) {
  assertTrue(this->state == TRUE_TYPE_ASSET_STATE_READY);

  // Try and find an existing texture that matches this style
  auto it = this->textureByStyle.find(style);
  TrueTypeFaceTexture *texture = nullptr;

  if(it == this->textureByStyle.end()) {
    // Does not exist, Find asset style
    auto itAssetStyle = this->assetStyles.begin();
    while(itAssetStyle != this->assetStyles.end()) {
      if((*itAssetStyle).style == style.style) {
        // Found
        break;
      }
      ++itAssetStyle;
    }
    assertTrue(itAssetStyle != this->assetStyles.end());

    // Create and read buffer.
    uint8_t *dataBuffer = (uint8_t*)memoryAllocate(sizeof(uint8_t) * itAssetStyle->dataSize);
    this->loader.rewind();
    this->loader.setPosition(itAssetStyle->dataOffset);
    auto read = this->loader.read(dataBuffer, itAssetStyle->dataSize);
    assertTrue(read == itAssetStyle->dataSize);

    // Create the face
    FT_Face face;
    auto ret = FT_New_Memory_Face(this->fontLibrary, (FT_Byte*)dataBuffer, itAssetStyle->dataSize, 0, &face);
    assertTrue(ret == 0);
    texture = new TrueTypeFaceTexture(face, style);
    memoryFree(dataBuffer);

    this->textures.push_back(texture);
    this->textureByStyle[style] = texture;
  } else {
    // Exists
    texture = it->second;
  }

  auto lock = texture->locks.createLock();
  this->textureByLock[lock] = texture;
  return lock;
}

TrueTypeFaceTexture * TrueTypeAsset::getTexture(usagelockid_t id) {
  auto it = this->textureByLock.find(id);
  assertTrue(it != this->textureByLock.end());
  return it->second;
}

void TrueTypeAsset::unlock(usagelockid_t id) {
  auto it = this->textureByLock.find(id);
  assertTrue(it != this->textureByLock.end());
  it->second->locks.removeLock(id);
  this->textureByLock.erase(it);
}

TrueTypeAsset::~TrueTypeAsset() {
  auto it = this->textures.begin();
  while(it != this->textures.end()) {
    delete (*it);
    it++;
  }

  FT_Done_FreeType(this->fontLibrary);
}